/*  $Id: ncbi_dblb.c 128770 2008-05-24 18:27:38Z lavr $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author:  Anton Lavrentiev
 *
 * File Description:
 *   Get DB service name via LB service name
 *
 *   NOTE:  This is an internal NCBI-only API not for export to third parties.
 *
 *   NOTE:  Non-UNIX platforms may experience some lack of functionality.
 *
 */

#include "../ncbi_ansi_ext.h"
#include "../ncbi_servicep.h"
#include <connect/ext/ncbi_dblb.h>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>


const char* DBLB_GetServerNameEx(const char* lb_name,
                                 char*       server_name_buf,
                                 size_t      server_name_buflen,
                                 const char* const skip_servers[],
                                 char*       errmsg_buf,
                                 size_t      errmsg_buflen)
{
    EDBLB_Status result;
    const char* retval =
        DBLB_GetServer(lb_name, 0/*p.host*/, 0/*p.port*/, 0.0/*preference*/,
                       skip_servers, 0/*host*/, 0/*port*/,
                       server_name_buf, server_name_buflen,
                       &result);
    if (errmsg_buf) {
        const char* errstr;
        assert(errmsg_buflen);
        errstr = DBLB_StatusStr(result);
        strncpy0(errmsg_buf,
                 errstr ? errstr : "Unknown error", errmsg_buflen - 1);
    }
    return retval;
}


const char* DBLB_GetServerName(const char* lb_name,
                               char*       server_name_buf,
                               size_t      server_name_buflen,
                               char*       errmsg_buf,
                               size_t      errmsg_buflen)
{
    return DBLB_GetServerNameEx(lb_name,
                                server_name_buf, server_name_buflen,
                                0, errmsg_buf, errmsg_buflen);
}


const char* DBLB_StatusStr(EDBLB_Status status)
{
    static const char* s_StatusStr[] = {
        "",
        "Bad service name",
        "Service not found",
        "No DNS entry found",
        "Service is down"
    };
    size_t i = (size_t) status;
    return i > sizeof(s_StatusStr)/sizeof(s_StatusStr[0]) ? 0 : s_StatusStr[i];
}


const char* DBLB_GetServerEx(const char*         lb_name,
                             unsigned int        preferred_host,
                             unsigned short      preferred_port,
                             double              preference,
                             const SConnNetInfo* net_info,
                             const char* const   skip_servers[],
                             unsigned int*       host,
                             unsigned short*     port,
                             char*               server_name_buf,
                             size_t              server_name_buflen,
                             EDBLB_Status*       result)
{
    static const char kPrefix[] = "DB_IP__"; 
    int/*bool*/ failed = 0/*false*/;
    size_t len, n, n_skip = 0;
    SSERV_Info** skip = 0;
    unsigned int   x_host;
    unsigned short x_port;
    SSERV_Info* info;
    EDBLB_Status x;
    const char* c;

    if (server_name_buf) {
        assert(server_name_buflen);
        server_name_buf[0] = '\0';
    }
    if (host)
        *host = 0;
    if (port)
        *port = 0;
    if (!lb_name  ||  !*lb_name) {
        if (result)
            *result = eDBLB_BadName;
        return 0;
    }
    if (result)
        *result = eDBLB_Success;
    else
        result = &x;

    if (skip_servers) {
        for (n = 0;  skip_servers[n];  n++)
            ;
        if ((skip = (SSERV_Info**) calloc(n, sizeof(*skip))) != 0) {
            for (n = 0;  skip_servers[n];  n++) {
                const char* server = skip_servers[n];
                SSERV_Info* info;
                if (!(len = strlen(server))) {
                    continue;
                }
                if (strncasecmp(server, kPrefix, sizeof(kPrefix)-1) == 0
                    &&  isdigit((unsigned char) server[sizeof(kPrefix)-1])) {
                    c = strstr(server + sizeof(kPrefix)-1, "__");
                    if (c) {
                        size_t i = (size_t)(c - server) - (sizeof(kPrefix)-1);
                        char* temp = strdup(server + sizeof(kPrefix)-1);
                        if (temp) {
                            char* s = temp + i;
                            *s++ = ':';
                            memmove(s, s + 1, strlen(s + 1) + 1);
                            server = temp;
                            while (++temp < s) {
                                if (*temp == '_')
                                    *temp =  '.';
                            }
                            len -= sizeof(kPrefix);
                        }
                    }
                }
                if (SOCK_StringToHostPort(server, &x_host, &x_port) !=
                    server + len) {
                    if (!(info = SERV_GetInfo(skip_servers[n], fSERV_Dns |
                                              fSERV_Promiscuous, 0,net_info))){
                        SSERV_Info* temp = SERV_CreateDnsInfo(0);
                        info = SERV_CopyInfoEx(temp, skip_servers[n]);
                        free(temp);
                    }
                } else if (server != skip_servers[n]) {
                    info = SERV_CreateStandaloneInfo(x_host, x_port);
                } else if ((info = SERV_CreateDnsInfo(x_host)) != 0) {
                    info->port = x_port;
                }
                if (info) {
                    skip[n_skip++] = info;
                }
                if (server != skip_servers[n]) {
                    free((void*) server);
                }
            }
        }
    }

    if (!(info = SERV_GetInfoP(lb_name, fSERV_ReverseDns | fSERV_Standalone,
                               preferred_host, preferred_port,
                               preference <=   0.0 ?  0.0 :
                               preference >= 100.0 ? -1.0 : preference,
                               net_info, skip, n_skip, 0/*not external*/,
                               0, 0, 0))) { /* NCBI_FAKE_WARNING: GCC */
        if (n_skip  &&  (x_host = SOCK_gethostbyname(lb_name)) != 0) {
            for (n = 0;  n < n_skip;  n++) {
                if (x_host == skip[n]->host) {
                    failed = 1/*true*/;
                    break;
                }
            }
        }
        if (!failed  &&  skip_servers) {
            for (n = 0;  (c = skip_servers[n]) != 0;  n++) {
                if (strcasecmp(c, lb_name) == 0) {
                    failed = 1/*true*/;
                    break;
                }
            }
        }
        if (!failed  &&  server_name_buf)
            strncpy0(server_name_buf, lb_name, server_name_buflen - 1);
        *result = eDBLB_NotFound;
    } else {
        if (info->type != fSERV_Dns) {
            char* s, buf[80];
            x_host = info->host;
            x_port = info->port;
            strncpy0(buf, kPrefix, sizeof(buf) - 1);
            SOCK_HostPortToString(x_host, x_port, buf + sizeof(kPrefix)-1,
                                  sizeof(buf) - sizeof(kPrefix));
            len = strlen(buf);
            if ((s = strchr(buf, ':')) != 0)
                memmove(s + 1, s, strlen(s) + 1);
            for (n = 0;  n < len;  n++) {
                if (buf[n] == '.'  ||  buf[n] == ':')
                    buf[n] = '_';
            }
            if (server_name_buf)
                strncpy0(server_name_buf, buf, server_name_buflen - 1);
            *result = eDBLB_NoDNSEntry;
        } else if (info->host) {
            x_host = info->host;
            x_port = info->port;
            c = SERV_NameOfInfo(info);
            assert(c);
            if (server_name_buf)
                strncpy0(server_name_buf, c, server_name_buflen - 1);
        } else {
            failed = 1/*true*/;
            *result = eDBLB_ServiceDown;
        }
        free(info);
        if (!failed) {
            if (host)
                *host = x_host;
            if (port)
                *port = x_port;
        }
    }

    for (n = 0;  n < n_skip;  n++)
        free(skip[n]);
    if (skip)
        free(skip);

    return failed ? 0 : (server_name_buf ? server_name_buf : lb_name);
}


const char* DBLB_GetServer(const char*       lb_name,
                           unsigned int      preferred_host,
                           unsigned short    preferred_port,
                           double            preference,
                           const char* const skip_servers[],
                           unsigned int*     host,
                           unsigned short*   port,
                           char*             server_name_buf,
                           size_t            server_name_buflen,
                           EDBLB_Status*     result)
{
    SConnNetInfo* net_info = ConnNetInfo_Create(lb_name);
    const char* retval = DBLB_GetServerEx(lb_name,
                                          preferred_host,
                                          preferred_port,
                                          preference,
                                          net_info,
                                          skip_servers,
                                          host,
                                          port,
                                          server_name_buf,
                                          server_name_buflen,
                                          result);
    ConnNetInfo_Destroy(net_info);
    return retval;
}
